/*
 *  SNAPSPIN: gives a N-body snapshot a spin around the Z-axis.
 *            also corrects potential and forces, if they are present
 *
 *	18-Aug-87	V1.0 created				PJT
 *	 8-Mar-88	V1.1 data history added			PJT
 *	 6-jun-88	V1.2 new filstruct			PJT
 *      25-apr-89       V1.3 get/put_snap , omega bug?          PJT
 *	 1-nov-90	V1.4 helpvec				PJT
 *	 2-may-92	V1.4a  usage 
 *	22-jul-92	V1.5 added rotcur=, sign= 		PJT
 *       6-oct-92       V1.6 fix potential, if present          PJT
 *       1-feb-93       V1.7 Phi/Acc corrections too            PJT
 *	mar-94 ansi
 */

#include <stdinc.h>
#include <getparam.h>
#include <vectmath.h>
#include <filestruct.h>
#include <spline.h>

#include <snapshot/snapshot.h>	
#include <snapshot/body.h>
#include <snapshot/get_snap.c>
#include <snapshot/put_snap.c>

string defv[] = {		/* DEFAULT INPUT PARAMETERS */
    "in=???\n		input file (snapshot)",
    "out=???\n		output filename",
    "omega=0.0\n	rotation (counterclock)",
    "rotcur=\n          Optional rotation curve spin (table: r,v)",
    "sign=1\n           Sign of rotcur",
    "times=all\n	times of snapshots to copy",
    "VERSION=1.7a\n	8-mar-94 PJT",	  
    NULL,
};

string usage="give a snapshot a spin around the Z-axis";

#define TIMEFUZZ	0.0001	/* tolerance in time comparisons */

double sqr();
real rotcur();

nemo_main()
{
    stream instr, outstr;
    real   tsnap, omega, omega2, sign, r;
    string headline = NULL;
    string times;
    Body *btab = NULL, *bp;
    int i, nbody, bits, nrscale, nvscale;
    bool Qrotcur, Qphi, Qacc;

    instr = stropen (getparam("in"), "r");
    omega=getdparam("omega");
    omega2 = omega*omega;
    Qrotcur = hasvalue("rotcur");
    sign = getdparam("sign");
    if (Qrotcur) {
        get_rotcur(getparam("rotcur"));
        if (omega != 0.0) warning("Ignoring spin omega=%g",omega);
    }
    outstr = stropen (getparam("out"), "w");
    times = getparam("times");    

    get_history(instr);
    put_history(outstr);					    					
    for(;;) {
    	get_history(instr);
        while (get_tag_ok(instr, HeadlineTag))
	    headline = get_string(instr, HeadlineTag);
        if (!get_tag_ok(instr, SnapShotTag))
	    break;			/* done with work in loop */
        get_snap(instr, &btab, &nbody, &tsnap, &bits);
        if ((bits & MassBit) == 0 && (bits & PhaseSpaceBit) == 0) {
            unlink(getparam("out"));
            error("essential data missing  bits = 0x%x\n", bits);
        }
        if ((bits & TimeBit) == 0)
            tsnap = 0.0;
        else if (!streq(times,"all") && !within(tsnap, times, TIMEFUZZ))
            continue;		/* however skip this snapshot */
        dprintf (1,"Snapshot time=%f spun\n",tsnap);
        Qphi = bits & PotentialBit;
        Qacc = bits & AccelerationBit;
        for (bp = btab; bp < btab+nbody; bp++) {
            if (Qrotcur) {
                r = sqrt(sqr(Pos(bp)[0])+sqr(Pos(bp)[1]));
                omega = rotcur(r)/r * sign;
                dprintf(1,"r=%g omega=%g v=%g\n",r,omega,r*omega);
                omega2 = omega * omega;
            }
            Vel(bp)[0] -= omega*Pos(bp)[1];          /*  vx = vx - y*omega */
            Vel(bp)[1] += omega*Pos(bp)[0];          /*  vy = vy + x*omega */
#if 1
            if (Qphi && !Qrotcur) 
                Phi(bp) += 0.5*omega2 * (sqr(Pos(bp)[0])+sqr(Pos(bp)[1]));
            if (Qacc && !Qrotcur) {
                Acc(bp)[0] += omega2 * Pos(bp)[0];
                Acc(bp)[1] += omega2 * Pos(bp)[1];
            }
#endif
        }
        bits = bits & (TimeBit | MassBit | PhaseSpaceBit);
        put_snap(outstr, &btab, &nbody, &tsnap, &bits);
    }
}

local real *rad, *vel, *coef;
local int nrad;

/* 
 * The rotation curve must be an ascii table(5NEMO) with 
 * radius in the first column and velocity in the second.
 * Currently one cannot use pipes yet, since the file_lines
 * would return a < 0 number.
 */

get_rotcur(name)
string name;
{
    stream instr;
    int  i,colnr[2], nmax;
    real *coldat[2];

    nmax = file_lines(name);
    if (nmax<1) error("No data to read from %s",name);
    rad = (real *) allocate(nmax*sizeof(real));
    vel = (real *) allocate(nmax*sizeof(real));
    coef= (real *) allocate(3*nmax*sizeof(real));   /* spline coeff's */

    colnr[0] = 1;       coldat[0] = rad;
    colnr[1] = 2;       coldat[1] = vel;

    instr = stropen(name,"r");
    nrad = get_atable(instr,2,colnr,coldat,nmax);
    strclose(instr);
    if (nrad > 2)
        spline(coef,rad,vel,nrad);
    else {
        warning("Could only read %d lines from %s - lin intpol",nrad,name);
        if (nrad<1) error("Not enuf lines (%d) for rotcur",nrad);
    }
    for(i=0; i<nrad; i++)
    	dprintf(1,"rad=%g vel=%g\n",rad[i],vel[i]);

}

real rotcur(r)
real r;
{
    if (r<=0.0)                 /* catch easy points */
        return 0.0;

    if (r<=rad[0])               /* catch case to linearly interpolate */
        return r*vel[0]/rad[0];

    if (r>=rad[nrad-1])          /* keplarian beyond last point */
        return vel[nrad-1]*sqrt(rad[nrad-1]/r);

    if (nrad==2) {          /* linear interpolate 2 points */
        return vel[0] + (r-rad[0])*(vel[1]-vel[0])/(rad[1]-rad[0]);
    }

    return seval(r,rad,vel,coef,nrad);      /* spline interpolation */
}

